$ !
$ ! Command file to load and startup CHAOSnet
$ !
$ ! The following privs are needed to execute this command proceedure
$ SET PROC/PRIV=(SYSNAM,CMKRNL,CMEXEC,DETACH,SETPRV,ALTPRI,PSWAPM)
$ !
$ ! Define logical names for where CHAOS software lives.
$ ! The format is @CHNAMES SYSTEM [<sysdev> [<srcdev> [<objdev>]]]
$ ! See CHNAMES.COM for details.
$ @SYS$MANAGER:CHNAMES SYSTEM
$ !
$ ! Define site's local time zone characteristics:
$ !
$ ! a) GMT_DSP is displacement west of GMT, in delta time format.
$ DEFINE/SYSTEM GMT_DSP  "0 05:00:00.0"
$ !
$ ! b) DST_SITE should be "ON" if this site is currently in a time zone
$ !    which is using daylight savings time.  Otherwise it should be "OFF" 
$ !    or not defined at all.
$ !*************************************************************************
$ !Note: This is a modification from previous releases - the DST_SITE logical
$ !	 should translate to "ON" ONLY WHEN DAYLIGHT SAVINGS TIME IS ACTUALLY
$ !	 IN EFFECT.  This sacrifice of user-friendliness is due to the 
$ !	 inclusion of the TIME server in the code executed by the NCP, which
$ !	 results in significant performance inprovement - the automatic
$ !	 evaluation of the current state of DST, based on date, may return
$ !	 in a future release. 
$ !*************************************************************************
$ DEFINE/SYSTEM DST_SITE "ON"
$ !
$ ! Define the defaults for the Remote Tape Facility:
$ ! If the name CHAOS$RTAPE_DEFAULTS exists, it has the format "drive/option..."
$ ! An option is one of ASSIST, NOASSIST, READ_CHECK, NOREAD_CHECK,
$ ! WRITE_CHECK, or NOWRITE_CHECK.
$ ! The "drive", if present, is the default drive to use if the user does not
$ ! specify one, or if he only specifies a unit number.  If only options are
$ ! present, then there is no default drive and users will have to specify a
$ ! drive with each use of the Remote Tape Facility.
$ DEFINE/SYSTEM CHAOS$RTAPE_DEFAULTS "_MTA0:/NOASSIST/READ_CHECK/WRITE_CHECK"
$ !
$ ! The following sets up the Chaosnet and Ethernet addresses.
$ ! CHNCP$CHAOSn should be the CHAOS network address of the n'th Chaos
$ ! interface.  CHNCP$INTERLANn should be the CHAOS network address of
$ ! the n'th Interlan Ethernet interface.  CHNCP$3COMn should be the
$ ! CHAOS network address of the n'th 3COM Ethernet interface (the
$ ! 3COM interface is not currently supported).  CHAOS$PRIMARY
$ ! should be the primary address for this host (as defined in the
$ ! host table) for normal operations, but can translate to the 
$ ! address of any of the node's interfaces.  The addresses are in octal.  
$ !**********************************************************************
$ ! Notes: If CHAOS$PRIMARY is not defined, the NCP will simply exit.
$ !	   At present only 1 INTERLAN interface per node is supported -
$ !	   running with multiple INTERLAN ethernet interfaces configured
$ !	   is likely to produce total confusion - in other words define
$ !	   ONLY CHNCP$INTERLAN0 or live dangerously.  This restriction 
$ !	   will be lifted in a future release.
$ !	   DO NOT confuse the CHAOS network address of an ethernet interface 
$ !	   with its hardware address.   All of the CHNCP$XXXn addresses 
$ !	   below are CHAOS NETWORK ADDRESSES (the octal addresses should
$ !	   consist of 2 eight-bit bytes, with the high order byte equal
$ !	   to the subnet number, and the low-order byte equal to the node
$ !	   number).  This 16-bit address WILL NOT CORRESPOND to the ethernet
$ !	   hardware address. 
$ !**********************************************************************
$ !
$ DEFINE/SYSTEM CHNCP$CHAOS0 17020
$ DEFINE/SYSTEM CHNCP$INTERLAN0 24434
$ DEFINE/SYSTEM CHNCP$PRIMARY 24434
$ !
$ ! The logical name CHNCP$NOBRIDGE controls whether this node will 
$ ! act as a network bridge, forwarding packets from one subnet to 
$ ! another. Mainly this equates to the broadcasting of routing table
$ ! info.  By default, any node which has more than one interface (as
$ ! defined by the previous logical names) will broadcast routing table
$ ! packets and act as a bridge.  To overide this function, define the
$ ! logical name CHNCP$NOBRIDGE to RESOLVE TO ANYTHING! This is key - the
$ ! NCP only checks for the existence of the logical name in the logical 
$ ! name table - the actual equivalence string is ignored (e.g.,
$ ! a) $DEFINE/SYSTEM CHNCP$NOBRIDGE "YES" and
$ ! b) $DEFINE/SYSTEM CHNCP$NOBRIDGE "NO" are equivalent to the NCP, and
$ ! either will disable bridging).
$ !
$! DEFINE/SYSTEM CHNCP$NOBRIDGE "BRIDGING_OFF"
$ !
$ ! The following loads the drivers for the CHAOSnet.  PC is the physical
$ ! CHAOSnet hardware, while CH is the logical CHAOSnet connection device.
$ ! One CH unit per connection is the rule, plus one for the NCP. MAXUNITS
$ ! can be adjusted (it is octal), and more or fewer CHAn units connected
$ ! as appropriate for your site.
$ !
$ ! The vector for the CH device is completely arbitrary; change it if needed.
$ ! The CSR and vector for the PC device correspond to the standard settings
$ ! for these, on the first unibus adapter.  If yours is an unusual
$ ! configuration, they will have be adjusted accordingly.
$ RUN SYS$SYSTEM:SYSGEN
LOAD CHAOS$SYSTEM:PCDRIVER
CONNECT PCA0 /ADAPTER=UB0/VEC=%O270/CSR=%O764140
LOAD CHAOS$SYSTEM:CHDRIVER
CONNECT CHA0 /NOADAPT/NUMVEC=1/MAXUNITS=20
CONNECT CHA1 /NOADAPT/NUMVEC=1
CONNECT CHA2 /NOADAPT/NUMVEC=1
CONNECT CHA3 /NOADAPT/NUMVEC=1
CONNECT CHA4 /NOADAPT/NUMVEC=1
CONNECT CHA5 /NOADAPT/NUMVEC=1
CONNECT CHA6 /NOADAPT/NUMVEC=1
CONNECT CHA8 /NOADAPT/NUMVEC=1
CONNECT CHA9 /NOADAPT/NUMVEC=1
CONNECT CHA10 /NOADAPT/NUMVEC=1
CONNECT CHA11 /NOADAPT/NUMVEC=1
CONNECT CHA12 /NOADAPT/NUMVEC=1
CONNECT CHA13 /NOADAPT/NUMVEC=1
CONNECT CHA14 /NOADAPT/NUMVEC=1
CONNECT CHA15 /NOADAPT/NUMVEC=1
EXIT
$ !
$ ! The following loads the Interlan driver for Chaosnet - see Interlan
$ ! documentation for CSR and VECTOR defaults.
$ !
$! $ RUN SYS$SYSTEM:SYSGEN
$! LOAD CHAOS$SYSTEM:NIDRIVER
$! CONNECT NIA0 /ADAPTER=UB0/CSR=%O764000/VECTOR=%O350/NUMVEC=2
$! EXIT
$ !
$ ! Connect up the PTYs for TELNET
$ RUN SYS$SYSTEM:SYSGEN
LOAD CHAOS$SYSTEM:PTDRIVER
LOAD CHAOS$SYSTEM:PYDRIVER
CONNECT PTA0 /NOADAPT/MAXUNITS=8
CONNECT PYA0 /NOADAPT/MAXUNITS=8
CONNECT PYA1 /NOADAPT/MAXUNITS=8
CONNECT PYA2 /NOADAPT/MAXUNITS=8
CONNECT PYA3 /NOADAPT/MAXUNITS=8
CONNECT PYA4 /NOADAPT/MAXUNITS=8
CONNECT PYA5 /NOADAPT/MAXUNITS=8
CONNECT PYA6 /NOADAPT/MAXUNITS=8
CONNECT PYA7 /NOADAPT/MAXUNITS=8
$ !
$ ! Install software that needs it
$ RUN SYS$SYSTEM:INSTALL
CHAOS$RFC:MAIL/SHARED/OPEN/HEADER_RESIDENT
CHAOS$RFC:NAME/SHARED/OPEN/HEADER_RESIDENT/PRIV=(CMEXEC,WORLD,SYSPRV)
CHAOS$RFC:SEND/SHARED/OPEN/HEADER_RESIDENT
CHAOS$RFC:SINK/SHARED/OPEN/HEADER_RESIDENT
CHAOS$RFC:FILE/SHARED/OPEN/HEADER_RESIDENT
CHAOS$RFC:TELNET/SHARED/OPEN/HEADER_RESIDENT/PRIV=(PHY_IO,TMPMBX,NETMBX)
CHAOS$SYSTEM:CFTP/SHARED/OPEN/HEADER_RESIDENT
CHAOS$SYSTEM:QSEND/SHARED/OPEN/HEADER_RESIDENT/PRIV=(SYSPRV,WORLD,OPER)
CHAOS$SYSTEM:STATUS/SHARED/OPEN/HEADER_RESIDENT
CHAOS$SYSTEM:SUPDUP/SHARED/OPEN/HEADER_RESIDENT
CHAOS$SYSTEM:TELUSR/SHARED/OPEN/HEADER_RESIDENT
$!
$!	Now start the NCP process
$!
$ @SYS$MANAGER:CHRESTART
